-- Module for saving/loading records

-- Save format (with \t as separator):
-- profile mapid name skin skincolor time

local SAVE_FILENAME = "multiplayer_records.txt"
local BACKUP_FILENAME = "multiplayer_records.1.txt"

local did_backup = false

local function tryReadRecord(arr, start, version)
    local ok, record = pcall(mpr.arrayToRecord, arr, start, version)
    
    if ok then return record end
end

local function tryWriteRecord(record)
    local ok, arr = pcall(mpr.recordToArray, record, true)
    
    if not ok then return end
    
    local ok, s = pcall(table.concat, arr, '\t')
    
    if ok then return s end
end

local function readRecords(file, only_map, only_profile, version)
    local linenum = 1
    local read = 0

    for line in file:lines() do
        linenum = linenum + 1

        -- Stolen from leaderboard
        local t = {}

        for word in (line.."\t"):gmatch("(.-)\t") do
            table.insert(t, word)
        end

        local profile = t[1]
        local mapid = tonumber(t[2])
        local record = tryReadRecord(t, 3, version)
        
        if not record then
            mpr.log("\130WARNING:\128 Malformed record at %s:%d", SAVE_FILENAME, linenum)
        else
            if (not only_map or mapid == only_map) and (not only_profile or profile == only_profile) then
                read = read + 1
                mpr.setRecord(mapid, record, profile)
            end
        end
    end
    
    mpr.log("Read %d/%d records (%d ignored lines)", read, linenum-1, linenum-1-read)
end

local function loadRecords(fn, map, profile)
    if not isserver then return end

    local file = io.open(fn or SAVE_FILENAME, "r")

    if not file then return end

    local version = tonumber(file:read())
    
	-- Now when i think about it, this isn't really recoverable
    if not version then
        mpr.log("\133ERROR:\128 %s: unknown version", fn or SAVE_FILENAME)
		return
    end
    
	mpr.log("Loading %s using format version '%d'", fn or SAVE_FILENAME, version)
	readRecords(file, map, profile, version)
    
    file:close()
end

local function saveRecords(fn)
    if not isserver then return end
    
    if not (fn and fn ~= SAVE_FILENAME) and not did_backup then
        mpr.log("Creating backup file %s", BACKUP_FILENAME)
        
        local orig = io.open(SAVE_FILENAME, "r")
        
        if orig then
            local bak = io.open(BACKUP_FILENAME, 'w')
            
            if bak then
                bak:write(orig:read("*a"))
                bak:close()
            else
                mpr.log("\130WARNING:\128 Couldn't open %s to back up records", BACKUP_FILENAME)
            end
            
            orig:close()
        else
            mpr.log("Nothing to back up")
        end
        
        did_backup = true
    end

    local file = io.open(fn or SAVE_FILENAME, "w")

    if not file then
        mpr.log("\130WARNING:\128 Couldn't open %s to save records", fn or SAVE_FILENAME)
        return
    end

    file:setvbuf("line")
    
    file:write(MPR_VERSION, '\n')
    
    local times = mpr.getTimesTable()
    
    local wrote = 0
    local records_count = 0

    for profile, records in pairs(times) do
        for mapid, record in pairs(records) do
            records_count = records_count + 1
            
            local s = tryWriteRecord(record)
            
            if not s then
                mpr.log("\130WARNING:\128 Failed to save record for map %s (#%d) in profile %s", G_BuildMapName(mapid), mapid, profile)
                mpr.log("Record dump (only mpr-relevant fields):")
                mpr.log(mpr.dumpRecord(record))
            else
                file:write(profile, '\t', mapid, '\t', s, '\n')
                wrote = wrote + 1
            end
        end
    end
    
    mpr.log("Wrote %d/%d records to %s (%d bad records)", wrote, records_count, fn or SAVE_FILENAME, records_count-wrote)
    
    file:close()
end

mpr.addHook("PreMapLoad", function()
    loadRecords()
end)

mpr.addHook("NewRecord", function()
    saveRecords()
end)

COM_AddCommand("mpr_io_save", function(p, fn)
    local fn = fn or SAVE_FILENAME

    CONS_Printf(p, "Force-saving records to "..fn)
    mpr.log("mpr_io_save: force-saving records to %s", fn)
    saveRecords(fn)
end, COM_ADMIN)

mpr.io = {
    loadRecords = loadRecords,
    saveRecords = saveRecords,
}